import random
import numpy as np
import pandas as pd
import igraph as ig
from numpy.random import uniform

class GeneRegulatoryNetwork:
    
    def create_grn(self):
        # load the GRN and extract the transcription factor network
        if self.verbose:
            print ("Cleaning Biological Network...", flush=True)
            
        self.grn = self.get_grn()
        tf_net = self.extract_tf_net()
        
        if tf_net.vcount() <= self.num_tfs:
            raise Exception("Number of Tfs is Larger than the GRN Provided...", flush=True)
        
        # sample tfs and e-genes
        if self.verbose:
            print ("Sampling TFs and E-Genes Network...", flush=True)
       
        tfs, tf_net = self.sample_tfs(tf_net)
        self.tf_egene_net = self.sample_egenes(tfs, tf_net)
    
    def get_grn(self):
        grn = pd.read_csv(self.grn_fp).set_index('row_names')
        edgelist = {'source':[], 'target':[], 'edge_weight':[]}

        for rowname, row in grn.iterrows():
            for colname in row.index:
                weight = row.loc[colname]

                if weight == 0:
                    continue

                edgelist['target'].append(colname)
                edgelist['source'].append(rowname)
                edgelist['edge_weight'].append(weight)

        edgelist = pd.DataFrame(edgelist)
        grn = ig.Graph.DataFrame(edgelist)
        return grn
    
    def extract_tf_net(self):
        tf_nodes = []
        nodes = [node.index for node in self.grn.vs]
        degrees = self.grn.degree(mode='out')

        for node_i, node in enumerate(nodes):
            if degrees[node_i] == 0:
                continue

            tf_nodes.append(node)

        tf_net = self.grn.induced_subgraph(tf_nodes)
        tf_net = self.clean_tf_net(tf_net)
        return tf_net
    
    def sample_tfs(self, tf_net):
        tfs = self.modular_sampling(tf_net, self.num_tfs)
        tf_net = tf_net.induced_subgraph(tfs)
        tf_net = self.rename_nodes(tf_net, tfs, 'TF')
        
        return tfs, tf_net
        
    def sample_egenes(self, tfs, tf_net):
        neighbors = self.get_neighbors(self.grn, tfs, 'out')
        tf_degree_net = self.grn.induced_subgraph(neighbors + tfs)
        tf_degree_net = self.rename_nodes(tf_degree_net, tfs, 'TF')
        df = self.get_edgelist_df(tf_degree_net)
        df = df.loc[df['source'].str.contains('TF_')]
        regs, probs, nrows = [], [], df.shape[0]
        
        for reg in df.source.unique():
            prob = df.loc[df.source == reg].shape[0] / nrows
            
            regs.append(reg)
            probs.append(prob)
        
        df = self.get_edgelist_df(tf_net)
        egenes = [f'E-Gene_{i}' for i in range(self.num_egenes)]
        new_regs = random.choices(regs, weights=probs, k=self.num_egenes)
        df = pd.concat([df, pd.DataFrame({'source': new_regs, 'target': egenes, 'edge_weight': uniform(size=self.num_egenes)})])
        return ig.Graph.DataFrame(df)
    
    def clean_tf_net(self, tf_net):
        graphs = tf_net.decompose(mode="WEAK")
        graph_edges = [graph.ecount() for graph in graphs]
        most_edges = max(graph_edges)

        for index, edge_count in enumerate(graph_edges):

            if edge_count == most_edges:
                tf_net = graphs[index]

        return tf_net


import os
import math
import random
import numpy as np
import pandas as pd
import scanpy as sc
from tqdm import tqdm
from multiprocessing import Pool, RLock
from numpy.random import uniform, multinomial

class Sampling:
#     need to look into using target genes
    def sample(self, sim_meta=None, res_dir=None, ncells=10000, lambda_ls=0, cache=True, pop_fp=None, targeted_transcriptome=None, map_ls=True):
        
        if res_dir is None:
            res_dir = self.results_dir
            
        if sim_meta is None:
            sim_meta = self.sim_meta
        
        print (f"Simulation: {self.network_name} Sampling Cells...", flush=True)
        cells_meta, gene_expr = self.sampling_cells(sim_meta, ncells, res_dir, targeted_transcriptome)
        
        print (f"Simulation: {self.network_name} Sampling Molecules...", flush=True)
        gene_expr, lib_sizes = self.sampling_molecules(gene_expr, map_ls, lambda_ls, pop_fp)
        cells_meta = self.clean_cells_metadata(cells_meta, lib_sizes)
        
        if cache:
            print (f"Simulation: {self.network_name} Caching....", flush=True)
            cells_meta.to_csv(os.path.join(self.metadata_dir, 'cells_metadata.csv.bz2'), compression='bz2', index=False)
            gene_expr.to_csv(os.path.join(self.metadata_dir, 'gene_expression.csv.bz2'), compression='bz2', index=False)
        
        return cells_meta, gene_expr
        
    def sampling_cells(self, sim_meta, ncells, sim_dir, targeted_transcriptome):
        meta_fp = os.path.join(sim_dir, 'cell_metadata.csv.gz')
        sim_fp = os.path.join(sim_dir, 'simulated_counts.csv.gz')
        
        cells_meta = pd.read_csv(meta_fp)\
                       .reset_index()\
                       .rename(columns={"index": "cell_i"})

        if ncells > cells_meta.shape[0]:
            raise Exception(f"Simulation: {self.network_name} Number of cells requested is greater than the number of cells simulated. Sample fewer cells...")
        
        cells_meta = self.sample_cells_per_grna(sim_meta, cells_meta, ncells)
        gene_expr = self.load_cells(cells_meta, sim_fp, targeted_transcriptome)
        
        return cells_meta, gene_expr
    
    def sampling_molecules(self, gene_expr, map_ls, lambda_ls, pop_fp=None):
        
        if pop_fp is None:
            pop = sc.read_loom(self.pop_fp)
        else:
            pop = sc.read_loom(pop_fp)
        
        realcounts = pop.X.toarray()
        cell_umis = pop.obs.total_counts.values
        
        simcounts_cpm = self.calc_cpm(realcounts, gene_expr)
        lib_size = self.calc_library_size(cell_umis, gene_expr, map_ls, lambda_ls)
        downsampled_counts = self.downsampling(simcounts_cpm, lib_size)
        gene_expr = pd.DataFrame(downsampled_counts, columns=gene_expr.columns, dtype=np.int16)
        
        return gene_expr, lib_size
    
    def clean_cells_metadata(self, meta, lib_sizes):
        meta['lib_size'] = lib_sizes
        meta['target_gene'] = meta['sim_name'].apply(lambda x: x.split('-grna')[0])
        
        if self.crispr_type == 'knockout':
            meta['is_cell_perturbed'] = meta['sim_name'].apply(lambda x: x.split('_')[-1])
            meta.loc[meta.target_gene == self.ctrl_label, 'is_cell_perturbed'] = self.ctrl_label
        else:
            meta['is_cell_perturbed'] = 'PRT'
            meta.loc[meta.target_gene == self.ctrl_label, 'is_cell_perturbed'] = self.ctrl_label
        
        meta = meta.reset_index(drop=True)
        return meta
    
    def load_cells(self, cells_meta, sim_fp, targeted_transcriptome):
        df = pd.read_csv(sim_fp, dtype=np.int16)
        df = df.iloc[cells_meta.cell_i.values]
        
        if not targeted_transcriptome is None and len(targeted_transcriptome) != 0:
            targeted_transcriptome = "mol_mrna_" + targeted_transcriptome
            df = df[targeted_transcriptome]
        
        return df
        
    def calc_library_size(self, cell_umis, sim_counts, map_ls, lambda_ls):
        nlibs = sim_counts.shape[0]
        
        if map_ls and lambda_lib_size == 0:
            # sampling library
            sim_probs = uniform(size=nlibs).astype(np.float16)
            lib_size = np.around(np.quantile(cell_umis, sim_probs))
        
        elif lambda_ls != 0:
            lib_size = np.random.poisson(lambda_ls, size=nlibs)
            
        else:
            lib_size = sim_counts.sum(axis=1).values
        
        return lib_size
    
    def calc_cpm(self, realcount, sim_counts):
        realcount_ls = np.sum(realcount, axis=1).reshape(-1, 1)
        sim_counts_ls = sim_counts.sum(axis=1).values.reshape(-1, 1)
        
        realcount_cpm = realcount / realcount_ls
        sim_counts_cpm = sim_counts / sim_counts_ls
        
        if self.map_reference_cpm:
            # sort sim counts data via least to greatest
            sim_shape = sim_counts.shape
            realcount_cpm = realcount_cpm.flatten()
            sim_cpm_size = sim_shape[0] * sim_shape[1]
            
            realcount_cpm = realcount_cpm[realcount_cpm != 0]
            
            print ('Sampling Probabilities...', flush=True)
            sim_probs = uniform(size=sim_cpm_size).astype(np.float16)
            
            print ('Fetching Quantiles...', flush=True)
            sim_counts_cpm = np.quantile(realcount_cpm.astype(np.float16), sim_probs).astype(np.float16).reshape(sim_shape)
            
            print ('Reshape into Expression Matrix...', flush=True)
            sim_counts_cpm = sim_counts_cpm / np.sum(sim_counts_cpm, axis=1).astype(np.float16).reshape(-1, 1)
        
        return sim_counts_cpm
    
    def downsampling(self, sim_counts_cpm, lib_sizes):
        print ('Downsampling Counts...', flush=True)
        
        for index, lib_size in tqdm(enumerate(lib_sizes)):
            gene_val = sim_counts_cpm[index, ]
            gene_val = gene_val.astype(np.float64)
            gene_val = gene_val / sum(gene_val)
            
            gene_expr = multinomial(lib_size, gene_val).astype(np.int16)
            sim_counts_cpm[index, ] = gene_expr        
        
        return sim_counts_cpm
    
    def sample_cells_per_grna(self, sim_meta, cells_meta, ncells):
        sampled_cells = []
        grnas = sim_meta.grna.unique()
        ncells_per_grna = math.ceil(ncells / len(grnas))
        
        for row_i in range(sim_meta.shape[0]):
            row = sim_meta.iloc[row_i]
            nsamples = int(math.ceil(ncells_per_grna * row.sample_percent))
            
            if nsamples == 0:
                continue
            
            cells_meta.loc[cells_meta.sim_name == row.sim_name, 'grna'] = row.grna
            sim_cells = cells_meta.loc[cells_meta.sim_name == row.sim_name, 'cell_i'].values
            sim_cells = list(sim_cells)
            
            if len(sim_cells) < nsamples:
                print ("changing ncells_per_grna...")
                nsamples = len(sim_cells)            

            sampled = random.sample(sim_cells, k=nsamples)
            sampled_cells = sampled_cells + sampled
        
        cells_meta = cells_meta.iloc[sampled_cells]
        return cells_meta



import os
import math
import random
import numpy as np
import pandas as pd
from numpy.random import uniform
from scipy.stats import truncnorm

class Kinetics:
    """I am concerned about the basal levels/effects. Check what previous basal levels are like from the R code."""
        
    def initialize_kinetics(self):
        self.sample_rates()
        self.sample_interactions()
        self.calc_dissociation()
        
        grn = self.feature_network.loc[~self.feature_network.kinase_edge, ]
        kinases = self.feature_network.loc[self.feature_network.kinase_edge, ]
        
        # calculate for GRN
        self.calc_basal_activity(grn)
        self.calc_effects_sums(grn)
        self.calc_base_activity()
        
        # calculate for phosphorylation
        self.calc_basal_activity(kinases, "kinase_basal")
        self.calc_effects_sums(kinases, "kinase_effects_sums")
        self.calc_base_activity("kinase_base_activity", "kinase_basal", "kinase_effects_sums")
        
        self.feature_network = self.feature_network.drop(columns=['feature_id'])
        print ("Initialized Kinetics...")
        
        if self.cache_network:
            self.feature_info.to_parquet(os.path.join(self.metadata_dir, 'feature_info.parquet'), compression='brotli')
            self.feature_network.to_parquet(os.path.join(self.metadata_dir, 'feature_network.parquet'), compression='brotli')
    
    def sample_rates(self):
        finfo = self.feature_info
        nrows = len(self.feature_info)
        nphospho = len(self.phosphorylated)
        finfo['transcription_rate'] = np.nan
        
        finfo['effects_sums'] = 0
        finfo['independence'] = 1
        finfo['base_activity'] = 0
        finfo['kinase_effects_sums'] = 0
        finfo['kinase_base_activity'] = 0
        finfo['phosphorylation_rate'] = 0
        finfo['dephosphorylation_rate'] = 0
        finfo['mrna_halflife'] = uniform(size=nrows,  high=5, low=2.5)
        finfo['protein_halflife'] = uniform(size=nrows, high=10, low=5)
        finfo['transcription_rate'] = uniform(size=nrows, high=50, low=20)
        finfo['translation_rate'] = uniform(size=nrows, high=150, low=100)
        finfo['splicing_rate'] = math.log(2) / 2
        finfo['mrna_decay_rate'] = math.log(2) / self.feature_info['mrna_halflife']
        finfo['protein_decay_rate'] = math.log(2) / self.feature_info['protein_halflife']
        finfo.loc[finfo.is_phosphorylated, 'phosphorylation_rate'] = uniform(size=nphospho, high=20, low=10)
        finfo.loc[finfo.is_phosphorylated, 'dephosphorylation_rate'] = uniform(size=nphospho, high=20, low=10)
        
        self.feature_info = finfo
        
    def sample_interactions(self):
        
        nrows = len(self.feature_network)
        fnet = self.feature_network.copy(deep=True)
        tfs = self.feature_info.loc[self.feature_info.is_tf, 'feature_id'].values
        ntf_edges = fnet.loc[fnet['from'].isin(tfs)].shape[0]
        
        fnet['basal'] = 0
        fnet['effect'] = random.choices([-1, 1], weights=[0.25, 0.75], k=nrows)
        fnet['strength'] = 10 ** uniform(size=nrows, high=2, low=1)
        fnet['hill'] = truncnorm.rvs(1, 10, loc=2, scale=2, size=nrows)
        
        self.feature_network = fnet
    
    def calc_dissociation(self):
        tfs = self.feature_info.loc[self.feature_info.is_tf, 'feature_id'].values
        remove = ["max_premrna", "max_mrna", "max_protein", "dissociation", "k", "max_protein"]
        
        finfo, fnet = self.feature_info, self.feature_network
        fnet = fnet[[col for col in fnet.columns if not col in remove]]
        finfo = finfo[[col for col in finfo.columns if not col in remove]]
        
        finfo["max_premrna"] = finfo["transcription_rate"] / (finfo["mrna_decay_rate"] + finfo["splicing_rate"])
        finfo["max_mrna"] = finfo["splicing_rate"] / finfo["mrna_decay_rate"] * finfo["max_premrna"]
        finfo["max_protein"] = finfo["translation_rate"] / finfo["protein_decay_rate"] * finfo["max_mrna"] 
        
        fnet = fnet.merge(finfo[['feature_id', 'max_protein']], left_on='from', right_on='feature_id', how='left')
        
        fnet['feature_id'] = fnet['to']
        fnet['dissociation'] = fnet['max_protein'] / 2
        self.feature_info, self.feature_network = finfo, fnet
    
    def calc_basal_activity(self, network, basal_col="basal"):
        basal_df = {"feature_id": [], basal_col: []}
        
        for index, group in network[['feature_id', 'effect']].groupby(['feature_id']):
            effects = group.effect.unique()
            basal_df["feature_id"].append(index)

            if len(effects) == 2:
                basal_df[basal_col].append(0.5)
            elif effects[0] == -1:
                basal_df[basal_col].append(1)
            elif effects[0] == 1:
                basal_df[basal_col].append(0.0001)
            else:
                basal_df[basal_col].append(np.nan)

        basal_df = pd.DataFrame(basal_df)
        self.feature_info = self.feature_info.merge(basal_df, on='feature_id', how='left')
        self.feature_info.loc[self.feature_info[basal_col].isna(), basal_col] = 1
    
    def calc_effects_sums(self, network, effects_col="effects_sums"):
        # calculating effects sums
        finfo = self.feature_info
        
        for feature, group in network.groupby(['to']):
            group = group.loc[group.effect > 0]
            
            if group.shape[0] == 0:
                pos_effects_sums = 0
            else:
                pos_effects_sums = group['effect'].sum()
            
            finfo.loc[finfo.feature_id == feature, effects_col] = pos_effects_sums
        
        self.feature_info = finfo
    
    def calc_base_activity(self, base_active_col="base_activity", basal_col="basal", effects_col="effects_sums"):
        info = self.feature_info
        above_zero = info[effects_col] > 0
        equal_or_below_zero = info[effects_col] <= 0

        info.loc[equal_or_below_zero, base_active_col] = info.loc[equal_or_below_zero, basal_col]
        info.loc[above_zero, base_active_col] = info.loc[above_zero,
                                                         basal_col] - info.loc[above_zero,
                                                                               'independence'] ** info.loc[above_zero, effects_col]
        
        info.loc[info[base_active_col].isna(), base_active_col] = 1
        self.feature_info = info
        
    def sample_interaction_effects(self, row):
        
        if 'TF' in row['from'] and 'TF' in row['to']:
            return random.choices([-1, 1], weights=[0.1, 0.9], k=1)[0]
        else:
            return random.choices([-1, 1], weights=[0.25, 0.75], k=1)[0]